#include <algorithm>
#include <iostream>
#include <queue>
#include <string>
#include <vector>

namespace sorting {

struct HuffmanTreeNode {

  virtual void print(const std::string&) = 0;

  virtual void generateTable(std::vector<std::string>& table, const std::string& pref = "") = 0;
};

struct HuffmanLeafNode : public HuffmanTreeNode {
  size_t word_id;

  HuffmanLeafNode(size_t word_id) : word_id(word_id) {}

  void print(const std::string& s) override {
    std::cout << s << " : " << word_id << std::endl;
  }

  void generateTable(std::vector<std::string>& table, const std::string& pref = "") override {
    if (table.size() <= word_id) table.resize(word_id + 1);
    table[word_id] = pref;
  }
};

struct HuffmanTreeInternalNode : public HuffmanTreeNode {
  HuffmanTreeNode* left, *right;

  HuffmanTreeInternalNode(HuffmanTreeNode* left, HuffmanTreeNode* right) : left(left), right(right) {}

  void print(const std::string& s) override {
    left->print(s + "0");
    right->print(s + "1");
  }

  void generateTable(std::vector<std::string>& table, const std::string& pref = "") override {
    left->generateTable(table, pref + "0");
    right->generateTable(table, pref + "1");
  }

  ~HuffmanTreeInternalNode() {
    delete left;
    delete right;
  }
};

template<class T>
using min_priority_queue = std::priority_queue<T, std::vector<T>, std::greater<T>>;

HuffmanTreeNode* generateHuffmanTree(const std::vector<size_t>& frequency_table) {
  min_priority_queue<std::pair<size_t, HuffmanTreeNode*>> pq;
  for (size_t i = 0; i < frequency_table.size(); ++i) {
    pq.push({ frequency_table[i], new HuffmanLeafNode(i) });
  }
  while (pq.size() != 1) {
    auto tp1 = pq.top(); pq.pop();
    auto tp2 = pq.top(); pq.pop();

    std::cout << "Merging : " << tp1.first << " + " << tp2.first << std::endl;
    pq.push({ tp1.first + tp2.first, new HuffmanTreeInternalNode(tp1.second, tp2.second) });
  }

  return pq.top().second;
}

}

int main() {
  std::vector<size_t> freqs({ 20, 3, 6, 7, 8, 56 });

  sorting::HuffmanTreeNode* tree = sorting::generateHuffmanTree(freqs);
  std::vector<std::string> table;
  tree->generateTable(table);

  // std::vector<std::string> expected({ "00", "0100", "0101", "0110", "0111", "1" });
  // EXPECT_EQ(table, expected);
  delete tree;
  
  return 0;
}
